#define WIN32_LEAN_AND_MEAN
#include <stdio.h>

#include "avisynth.h"
#include "info.h"

const int		PriorityConversion[8] = {
				THREAD_PRIORITY_IDLE, THREAD_PRIORITY_LOWEST,
				THREAD_PRIORITY_BELOW_NORMAL, THREAD_PRIORITY_NORMAL,
				THREAD_PRIORITY_BELOW_NORMAL, THREAD_PRIORITY_HIGHEST,
				THREAD_PRIORITY_TIME_CRITICAL };

#define			MAX_TIMERS		6
				
typedef struct timer {
	__int64			timeinns;				//Total execution time
	__int64			temps;					//Last execution time
	unsigned int	count;					//Number of frames so far on output
	unsigned int	nrequests;				//last number of frames requested by input
	bool			seen_start, seen_stop;	//Corresponding start/stoptimer() present
	bool			full;					//Prints on screen
	bool			mb;						//To pop up a MessageBox on exit
	char			LogFileName[1024];		//Logfilename
	bool			HaveFN;					//Do we have a filename?
	int				nPriority;				//Priority of thread in StopTimer()
	char			Name[255];				//Name of the timer
	SYSTEMTIME		StartTime;				//To print in logile
	int				ypos;					//Position for StopTimer();
} MyTimer;

//Global - number of timers
int				TimerCount, TotalFull;
MyTimer			Timer[MAX_TIMERS];			//Should be sufficient, global to be shared

int dprintf(char* fmt, ...)
{
	char printString[5000];
	va_list argp;
	va_start(argp, fmt);
	vsprintf(printString, fmt, argp);
	va_end(argp);
    
	OutputDebugString(printString);
	return strlen(printString);
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpReserved)
{
	if (fdwReason == DLL_PROCESS_DETACH)
	{
		if (TimerCount != 0)
		{
			int		i;
			char	EndText[2048], EndMBText[2048];
			double	temps;
			bool	DispMB = false;
			FILE	*LogFile;

			sprintf(EndMBText, "Results:");
			for (i=0; i<TimerCount; i++)
			{
				temps = (Timer[i].timeinns/(__int64)1000)/(10.0*Timer[i].count);
				sprintf(EndText,
						"After %i frames for timer \"%s\", average time = %5.2f milliseconds, average speed = %4.2f fps\n",
						Timer[i].count, Timer[i].Name, temps, 1000.0/temps);
				if (temps != 0)
					dprintf("Kronos> %s\n", EndText);
				else OutputDebugString("Kronos> faster than NULL filter detected...");
	
				if (Timer[i].mb)
				{
					if (Timer[i].timeinns == 0)
						sprintf(EndMBText, "%s\nTimer %s: Faster than NULL", EndMBText, Timer[i].Name);
					else
						sprintf(EndMBText, "%s\nTimer %s: after %i frames\nAverage time = %5.2f milliseconds\nAverage speed = %4.2f fps\n",
								EndMBText, Timer[i].Name, Timer[i].count, temps, 1000.0/temps);
				}
				DispMB |= Timer[i].mb;

				if (Timer[i].HaveFN)
				{
					SYSTEMTIME	utc, local;

					LogFile = fopen(Timer[i].LogFileName, "at+");
					GetSystemTime(&utc);

					SystemTimeToTzSpecificLocalTime(NULL,&Timer[i].StartTime, &local);
					fprintf(LogFile, "Script started %.2i/%.2i/%.4i at %.2i:%.2i:%.2i.%.1i;",
							Timer[i].Name, local.wMonth, local.wDay, local.wYear, local.wHour, local.wMinute,
							local.wSecond, local.wMilliseconds);
					SystemTimeToTzSpecificLocalTime(NULL, &utc, &local);
					fprintf(LogFile, "Finished %.2i/%.2i/%.4i at %.2i:%.2i:%.2i.%.1i\n",
							local.wMonth, local.wDay, local.wYear, local.wHour, local.wMinute,
							local.wSecond, local.wMilliseconds);

					if (Timer[i].timeinns != 0) fprintf(LogFile,EndText);
					else fprintf(LogFile,"Faster than NULL filter detected\n");
					
					fprintf(LogFile,"\n----\n\n", EndText);
					fclose(LogFile);
				}
			}
			if (DispMB)	MessageBox(GetActiveWindow(), EndMBText, "Kronos timings", MB_OK);
		}
	}

	return TRUE;
}

typedef	struct param {
	PClip				Clip;
	PVideoFrame			PVF;
	int					Frame;
	IScriptEnvironment	*env;
} Param;

unsigned long __stdcall RequestFrame( LPVOID pJunk )
{
	Param *MypParam = (Param *)pJunk;
	MypParam->PVF = MypParam->Clip->GetFrame(MypParam->Frame, MypParam->env);
	return 0;
}


class Start : public GenericVideoFilter {
private:
	SECURITY_ATTRIBUTES saPipe;
	int					index;
public:
	Start(PClip _child, const char *tname, IScriptEnvironment* env) : GenericVideoFilter(_child)
	{
		int i;

		//Check if a name was provided
		if (tname == NULL)
			env->ThrowError("StartTimer: Give a name to this timer");
		
		//Check for the existence of the filter
		for (i=0; i<TimerCount-1; i++)
		{
			/* Check if one by that name already exists */
			if (lstrcmpi(Timer[i].Name,tname) == 0)
			{
				if (Timer[i].seen_stop)
					env->ThrowError("StartTimer: %s must be started before it's stopped...", tname);
				else
					env->ThrowError("StartTimer: %s was already initialized", tname);
			}
		}

		index = TimerCount++;

		//Last check
		if (TimerCount>MAX_TIMERS)
			env->ThrowError("StartTimer: more than %i timers initialized within the same script", MAX_TIMERS);

		//Now updates the initializaton of the counter
		Timer[index].seen_start = true;
		strcpy(Timer[index].Name, tname);

		saPipe.nLength = sizeof(SECURITY_ATTRIBUTES);
		saPipe.lpSecurityDescriptor = NULL;
		saPipe.bInheritHandle = true;
	}
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env)
	{
#if 0
		Param		MyParam;
		DWORD		dwThreadId = 1;
		HANDLE		hThread = NULL;
		FILETIME	CreationTime, ExitTime, KernelTime, UserTime;
		__int64		Kernel, User;

		MyParam.Frame = n;
		MyParam.env = env;
		MyParam.Clip = child;
		hThread =
			CreateThread(&saPipe, 0, RequestFrame, &MyParam,
						THREAD_QUERY_INFORMATION | SYNCHRONIZE | THREAD_SET_INFORMATION,
						&dwThreadId);
		if (hThread == NULL)
			env->ThrowError("Kronos: Couldn't start thread in start()");
		SetThreadPriority(hThread, Timer[index].nPriority);
		if (WaitForSingleObject(hThread, INFINITE) != WAIT_OBJECT_0)
			env->ThrowError("Kronos: thread aborted in start()");;
		if (!GetThreadTimes(hThread, &CreationTime, &ExitTime, &KernelTime, &UserTime))
			env->ThrowError("Kronos: Couldn't request times from thread in start()");
		if (CloseHandle(hThread) == NULL)
			env->ThrowError("Kronos: Couldn't close thread in start()");
		Kernel = ((__int64)KernelTime.dwHighDateTime<<32) +
				(__int64)KernelTime.dwLowDateTime;
		User = ((__int64)UserTime.dwHighDateTime<<32) +
				(__int64)UserTime.dwLowDateTime;
		Timer[index].temps += Kernel + User;		//units of 100ns
		Timer[index].nrequests++;
		return MyParam.PVF;
#else
		LARGE_INTEGER	frequency, count1, count2;
		QueryPerformanceCounter(&count1);
		QueryPerformanceFrequency(&frequency);
		PVideoFrame src = child->GetFrame(n, env);
		Timer[index].nrequests++;
		QueryPerformanceCounter(&count2);
		Timer[index].temps += count2.QuadPart - count1.QuadPart;
		return src;
#endif
	}
};

class Stop : public GenericVideoFilter {
private:
	char				Msg[512];
	int					ypos, index;
	SECURITY_ATTRIBUTES	saPipe;
protected:
	void				(* DrawStringOnScreen) (PVideoFrame &dst, int x, int y, const char *s);
public:
	Stop(PClip _child, const char *tname, bool display, bool messagebox, const char *logfn,
		int pr, IScriptEnvironment* env) : GenericVideoFilter(_child)
	{
		int i;
		if (pr<0 || pr>6) env->ThrowError("StartTimer: %i isn't a valid priority", pr);

		//Check if a name was provided
		if (tname == NULL)
			env->ThrowError("StartTimer: Give a name to this timer");
		
		//Check for the existence of the timer
		for (i=0; i<TimerCount; i++)
		{
			/* Check if one by that name already exists */
			if (lstrcmpi(Timer[i].Name,tname) == 0)
			{
				//All good, finish initialization
				if (Timer[i].seen_start)
				{
					index = i;
					Timer[i].seen_stop = true;
					Timer[i].full = display;
					TotalFull += display;
					Timer[i].mb = messagebox;
					Timer[i].nPriority = PriorityConversion[pr];
					Timer[i].ypos = vi.height / 20 - 2*TotalFull;
					break;
				}
				else
					env->ThrowError("StopTimer: %s wasn't initialized yet", tname);
			}
		}
		if (logfn != NULL)
		{
			FILE		*LogFile;
			LogFile = fopen(logfn, "at+");
			if (LogFile == NULL)
				env->ThrowError("Kronos: couldn't open logfile \"%\"", logfn);
			Timer[i].HaveFN = true;
			strncpy(Timer[i].LogFileName, logfn, strlen(logfn));
			GetSystemTime(&Timer[i].StartTime);
			fclose(LogFile);
		}

		saPipe.nLength = sizeof(SECURITY_ATTRIBUTES);
		saPipe.lpSecurityDescriptor = NULL;
		saPipe.bInheritHandle = true;

		if (vi.IsYUY2())
			DrawStringOnScreen = DrawStringYUY2;
		else if (vi.IsRGB32())
			DrawStringOnScreen = DrawStringRGB32;
		else if (vi.IsYV12())
			DrawStringOnScreen = DrawString;
		else
		{
			MessageBox(NULL, "Display (full=true) not available in this colorspace\nDisplay is now disabled",
						"Kronos information", MB_OK | MB_ICONEXCLAMATION);
			Timer[index].full = false;
		}
	}

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env)
	{
#if 0
		Param		MyParam;
		DWORD		dwThreadId = 1;
		HANDLE		hThread = NULL;
		FILETIME	CreationTime, ExitTime, KernelTime, UserTime;
		__int64		Kernel, User;

		MyParam.Frame = n;
		MyParam.env = env;
		MyParam.Clip = child;
		Timer[index].temps = 0;
		Timer[index].nrequests = 0;
		hThread =
			CreateThread(&saPipe, 0, RequestFrame, &MyParam,
						THREAD_QUERY_INFORMATION | SYNCHRONIZE | THREAD_SET_INFORMATION,
						&dwThreadId);
		if (hThread == NULL)
			env->ThrowError("StopTimer: Couldn't start thread");
		SetThreadPriority(hThread, PriorityConversion[Timer[index].nPriority]);
		if (WaitForSingleObject(hThread, INFINITE) != WAIT_OBJECT_0)
			env->ThrowError("StopTimer: thread aborted");
		if (GetThreadTimes(hThread, &CreationTime, &ExitTime, &KernelTime, &UserTime) == NULL)
			env->ThrowError("StopTimer: Couldn't request times");
		if (CloseHandle(hThread) == NULL)
			env->ThrowError("StopTimer: Couldn't close thread");

		Kernel = ((__int64)KernelTime.dwHighDateTime<<32) +
				(__int64)KernelTime.dwLowDateTime;
		User = ((__int64)UserTime.dwHighDateTime<<32) +
				(__int64)UserTime.dwLowDateTime;
		Timer[index].count++;
		Timer[index].temps = User + Kernel - Timer[index].temps;
		Timer[index].timeinns += Timer[index].temps;
		Timer[index].temps /= (__int64)1000;

		if (Timer[index].full)
		{
			env->MakeWritable(&MyParam.PVF);
			sprintf(Msg,"Timer \"%s\", frame %i (%i in):", Timer[index].Name, n, Timer[index].nrequests);
			DrawStringOnScreen(MyParam.PVF, 0, Timer[index].ypos, Msg);
			sprintf(Msg,"%5.2f ms, avg. = %5.2f ms/%5.2f fps",
					Timer[index].temps/10.0 + 0.005,
					(Timer[index].timeinns/(__int64)1000)/(10.0*Timer[index].count) + 0.005,
					Timer[index].count*10000.0/(Timer[index].timeinns/(__int64)1000) + 0.005 );
			DrawStringOnScreen(MyParam.PVF, 0, Timer[index].ypos+1, Msg);
		}
		sprintf(Msg,"Kronos> Timer \"%s\", frame %i (%i in): %5.2f ms, avg. = %5.2f ms/%5.2f fps\n",
				Timer[index].Name, n, Timer[index].nrequests, Timer[index].temps/10.0 + 0.005,
				(Timer[index].timeinns/(__int64)1000)/(10.0*Timer[index].count) + 0.005,
				Timer[index].count*10000.0/(Timer[index].timeinns/(__int64)1000) + 0.005 );
		OutputDebugString(Msg);
		return MyParam.PVF;
#else
		LARGE_INTEGER	frequency, count1, count2;
		QueryPerformanceFrequency(&frequency);
		Timer[index].temps = 0;
		Timer[index].nrequests = 0;
		
		QueryPerformanceCounter(&count1);
		PVideoFrame src = child->GetFrame(n, env);
		QueryPerformanceCounter(&count2);

		Timer[index].count++;
		Timer[index].temps = count2.QuadPart - count1.QuadPart - Timer[index].temps;
		Timer[index].timeinns += Timer[index].temps;
		Timer[index].temps /= (__int64)1000;

		if (Timer[index].full)
		{
			env->MakeWritable(&src);
			sprintf(Msg,"Timer \"%s\", frame %i (%i in):", Timer[index].Name, n, Timer[index].nrequests);
			DrawStringOnScreen(src, 0, Timer[index].ypos, Msg);
			sprintf(Msg,"%5.2f ms, avg. = %5.2f ms/%5.2f fps",
					Timer[index].temps/10.0 + 0.005,
					(Timer[index].timeinns/(__int64)1000)/(10.0*Timer[index].count) + 0.005,
					Timer[index].count*10000.0/(Timer[index].timeinns/(__int64)1000) + 0.005 );
			DrawStringOnScreen(src, 0, Timer[index].ypos+1, Msg);
		}
		sprintf(Msg,"Kronos> Timer \"%s\", frame %i (%i in): %5.2f ms, avg. = %5.2f ms/%5.2f fps\n",
				Timer[index].Name, n, Timer[index].nrequests, Timer[index].temps/10.0 + 0.005,
				(Timer[index].timeinns/(__int64)1000)/(10.0*Timer[index].count) + 0.005,
				Timer[index].count*10000.0/(Timer[index].timeinns/(__int64)1000) + 0.005 );
		OutputDebugString(Msg);

		return src;
#endif
	}
};

AVSValue __cdecl Kronometer_Start(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	return new Start(
		args[0].AsClip(),		
		args[1].AsString(NULL), //NULL timer name -> bad
		env);
}

AVSValue __cdecl Kronometer_Stop(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	return new Stop(
		args[0].AsClip(),
		args[1].AsString(NULL), //NULL timer name -> bad
		args[2].AsBool(false),  //Don't print on screen
		args[3].AsBool(false),  //Don't display the MessageBox
		args[4].AsString(NULL),	//NULL name for log file -> no logging
		args[5].AsInt(3),		//Translated to THREAD_PRIORITY_NORMAL
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env)
{
	int i;

	TimerCount = 0;		//Simplest way to initialize it with dll call
	TotalFull = 0;

	//Pre-initialize timers
	for (i=0; i<MAX_TIMERS; i++)
	{
		Timer[i].full = Timer[i].mb = false;
		Timer[i].count = 0;
		Timer[i].nPriority = PriorityConversion[3];
		Timer[i].seen_start = Timer[i].seen_start = false;
		Timer[i].temps = Timer[i].timeinns = 0i64;
		Timer[i].HaveFN = false;
	}

	//Do the exports
	env->AddFunction("StopTimer", "c[name]s[full]b[mb]b[log]s[priority]i", Kronometer_Stop, 0);
	env->AddFunction("StartTimer", "c[name]s", Kronometer_Start, 0);
	return "'Kronos' let you have an estimation of block execution duration";
}
